import typing as t
from inspect import iscoroutinefunction

from jinja2.exceptions import TemplateNotFound
from jinja2.loaders import BaseLoader
from jinja2.loaders import ChoiceLoader as JinjaChoiceLoader
from jinja2.loaders import DictLoader as JinjaDictLoader
from jinja2.loaders import FileSystemLoader as JinjaFileSystemLoader
from jinja2.loaders import FunctionLoader as JinjaFunctionLoader
from jinja2.loaders import ModuleLoader as JinjaModuleLoader
from jinja2.loaders import PackageLoader as JinjaPackageLoader
from jinja2.loaders import PrefixLoader as JinjaPrefixLoader
from jinja2.utils import internalcode

if t.TYPE_CHECKING:
    from .environment import AsyncEnvironment, AsyncTemplate


class AsyncBaseLoader(BaseLoader):
    async def get_source_async(
        self, environment: "AsyncEnvironment", template: str
    ) -> tuple[str, str | None, t.Callable[[], t.Awaitable[bool] | bool] | None]:
        """Asynchronously get the template source, filename and reload helper
        for a template.
        It's passed the environment and template name and has to return a
        tuple in the form ``(source, filename, uptodate)`` or raise a
        `TemplateNotFound` error if it can't locate the template.

        The source part of the returned tuple must be the source of the
        template as a string. The filename should be the name of the
        file on the filesystem if it was loaded from there, otherwise
        ``None``. The filename is used by Python for the tracebacks
        if no loader extension is used.

        The last item in the tuple is the `uptodate` function.  If auto
        reloading is enabled it's always called to check if the template
        changed.  No arguments are passed so the function must store the
        old state somewhere (for example in a closure).  If it returns `False`
        the template will be reloaded.
        """
        return self.get_source(environment, template)

    async def list_templates_async(self) -> list[str]:
        """Asynchronously iterates over all templates.  If the loader does
        not support that it should raise a :exc:`TypeError` which is the
        default behavior.
        """
        return self.list_templates()

    async def load_async(
        self,
        environment: "AsyncEnvironment",
        name: str,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> "AsyncTemplate":
        """Asynchronously loads a template.  This method looks up the template
        in the cache or loads one by calling :meth:`get_source_async`.
        Subclasses should not override this method as loaders working on
        collections of other loaders (such as :class:`PrefixLoader` or
        :class:`ChoiceLoader`) will not call this method but
        `get_source_async` directly.
        """
        code = None
        if globals is None:
            globals = {}

        source, filename, uptodate = await self.get_source_async(environment, name)

        bcc = environment.bytecode_cache
        if bcc is not None:
            bucket = await bcc.get_bucket_async(environment, name, filename, source)
            code = bucket.code

        # if we don't have code so far (not cached, no longer up to
        # date) etc. we compile the template
        if code is None:
            code = environment.compile(source, name, filename)

        # if the bytecode cache is available and the bucket doesn't
        # have a code so far, we give the bucket the new code and put
        # it back to the bytecode cache.
        if bcc is not None and bucket.code is None:
            bucket.code = code
            await bcc.set_bucket_async(bucket)

        return environment.template_class.from_code(
            environment, code, globals, uptodate
        )


class FileSystemLoader(JinjaFileSystemLoader, AsyncBaseLoader):
    pass


class PackageLoader(JinjaPackageLoader, AsyncBaseLoader):
    pass


class DictLoader(JinjaDictLoader, AsyncBaseLoader):
    pass


class ModuleLoader(JinjaModuleLoader, AsyncBaseLoader):
    pass


class FunctionLoader(JinjaFunctionLoader, AsyncBaseLoader):
    load_func: t.Callable[
        [str],
        str
        | tuple[str, str | None, t.Callable[[], t.Awaitable[bool] | bool] | None]
        | None
        | t.Awaitable[
            str
            | tuple[str, str | None, t.Callable[[], t.Awaitable[bool] | bool] | None]
            | None
        ],
    ]

    async def get_source_async(
        self, environment: "AsyncEnvironment", template: str
    ) -> tuple[str, str | None, t.Callable[[], t.Awaitable[bool] | bool] | None]:
        if iscoroutinefunction(self.load_func):
            rv = await self.load_func(template)
        else:
            rv = self.load_func(template)

        if rv is None:
            raise TemplateNotFound(template)

        if isinstance(rv, str):
            return rv, None, None

        return rv


class PrefixLoader(JinjaPrefixLoader, AsyncBaseLoader):
    mapping: t.Mapping[str, AsyncBaseLoader]

    if t.TYPE_CHECKING:

        def get_loader(self, template: str) -> tuple[AsyncBaseLoader, str]: ...

    async def get_source_async(
        self, environment: "AsyncEnvironment", template: str
    ) -> tuple[str, str | None, t.Callable[[], t.Awaitable[bool] | bool] | None]:
        rv = self.get_loader(template)

        if rv is None:
            raise TemplateNotFound(template)

        loader, name = rv
        try:
            return await loader.get_source_async(environment, name)
        except TemplateNotFound as e:
            raise TemplateNotFound(template) from e

    @internalcode
    async def load_async(
        self,
        environment: "AsyncEnvironment",
        name: str,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> "AsyncTemplate":
        rv = self.get_loader(name)

        if rv is None:
            raise TemplateNotFound(name)

        loader, local_name = rv
        try:
            return await loader.load_async(environment, local_name, globals)
        except TemplateNotFound as e:
            raise TemplateNotFound(name) from e

    async def list_templates_async(self) -> list[str]:
        result = []
        for prefix, loader in self.mapping.items():
            for template in await loader.list_templates_async():
                result.append(prefix + self.delimiter + template)
        return result


class ChoiceLoader(JinjaChoiceLoader, AsyncBaseLoader):
    loaders: t.Sequence[AsyncBaseLoader]

    async def get_source_async(
        self, environment: "AsyncEnvironment", template: str
    ) -> tuple[str, str | None, t.Callable[[], t.Awaitable[bool] | bool] | None]:
        for loader in self.loaders:
            try:
                return await loader.get_source_async(environment, template)
            except TemplateNotFound:
                pass
        raise TemplateNotFound(template)

    @internalcode
    async def load_async(
        self,
        environment: "AsyncEnvironment",
        name: str,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> "AsyncTemplate":
        for loader in self.loaders:
            try:
                return await loader.load_async(environment, name, globals)
            except TemplateNotFound:
                pass
        raise TemplateNotFound(name)

    async def list_templates_async(self) -> list[str]:
        found = set()
        for loader in self.loaders:
            found.update(await loader.list_templates_async())
        return sorted(found)



import re

from jinja2.compiler import CodeGenerator as JinjaCodeGenerator

# Matches template lookup calls with optional assignment prefix or await prefix
# Groups: (1) assignment like "template = ", (2) await prefix, (3) method name
_TEMPLATE_CALL_RE = re.compile(
    r"((?:template|parent_template)\s*=\s*)?"
    r"(await\s+)?"
    r"environment\.(get_template|select_template|get_or_select_template)\("
)

# Matches the closing pattern for import statements ending with ).
# This handles: ", 'template_name')." from _import_common
_IMPORT_CLOSE_RE = re.compile(r"\)\.$")


class AsyncCodeGenerator(JinjaCodeGenerator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Track depth of import-style template calls that need extra closing paren
        self._import_paren_depth = 0

    def choose_async(self, async_value: str = "async ", sync_value: str = "") -> str:
        return async_value

    def _rewrite(self, s: str) -> str:
        def replace(m: re.Match) -> str:
            assign, await_prefix, method = m.groups()
            if assign:
                # Assignment pattern: template = environment.get_template(
                # -> template = await environment.get_template_async(
                return f"{assign}await environment.{method}_async("
            elif await_prefix:
                # Import pattern: await environment.get_template(
                # -> await (await environment.get_template_async(
                # Track that we need to close the extra paren later
                self._import_paren_depth += 1
                return f"await (await environment.{method}_async("
            else:
                # Bare call
                return f"await environment.{method}_async("

        s = _TEMPLATE_CALL_RE.sub(replace, s)

        # Handle closing pattern for import statements: "). at end of string
        # We need to add extra ) to close the (await ...) wrapper
        if self._import_paren_depth > 0 and _IMPORT_CLOSE_RE.search(s):
            s = _IMPORT_CLOSE_RE.sub(")).", s)
            self._import_paren_depth -= 1

        return s

    def write(self, x: str) -> None:
        super().write(self._rewrite(x))

    def writeline(self, x: str = "", node=None, extra=0) -> None:
        super().writeline(self._rewrite(x), node, extra)



from jinja2.sandbox import SandboxedEnvironment

from .environment import AsyncEnvironment


class AsyncSandboxedEnvironment(SandboxedEnvironment, AsyncEnvironment):
    """A sandboxed environment that supports async template loaders.
    The sandboxed environment works like the regular environment but
    tells the compiler to generate sandboxed code.  Additionally subclasses of
    this environment may override the methods that tell the runtime what
    attributes or functions are safe to access.

    If the template tries to access insecure code a :exc:`SecurityError` is
    raised.  However, also other exceptions may occur during the rendering so
    the caller has to ensure that all exceptions are caught.
    """

    sandboxed = True



import typing as t

from jinja2.bccache import Bucket, BytecodeCache
from jinja2.bccache import FileSystemBytecodeCache as JinjaFileSystemBytecodeCache
from jinja2.bccache import MemcachedBytecodeCache as JinjaMemcachedBytecodeCache

if t.TYPE_CHECKING:
    from .environment import AsyncEnvironment


class AsyncBytecodeCache(BytecodeCache):
    async def load_bytecode_async(self, bucket: Bucket) -> None:
        """Subclasses have to override this method to load bytecode into a
        bucket asynchronously.  If they are not able to find code in the
        cache for the bucket, it must not do anything.
        Unless overridden; this copies the behaviour of load_bytecode.
        """
        return self.load_bytecode(bucket)

    async def dump_bytecode_async(self, bucket: Bucket) -> None:
        """Subclasses have to override this method to write the bytecode
        from a bucket back to the cache asynchronously.  If it is unable to
        do so it must not fail silently but raise an exception.
        Unless overridden; this copies the behaviour of dump_bytecode.
        """
        return self.dump_bytecode(bucket)

    async def get_bucket_async(
        self,
        environment: "AsyncEnvironment",
        name: str,
        filename: str | None,
        source: str,
    ) -> Bucket:
        """Asynchronously return a cache bucket for the given template.  All
        arguments are mandatory but filename may be `None`.
        """
        key = self.get_cache_key(name, filename)
        checksum = self.get_source_checksum(source)
        bucket = Bucket(environment, key, checksum)
        await self.load_bytecode_async(bucket)
        return bucket

    async def set_bucket_async(self, bucket: Bucket) -> None:
        """Asynchronously put the bucket into the cache."""
        await self.dump_bytecode_async(bucket)


class FileSystemBytecodeCache(JinjaFileSystemBytecodeCache, AsyncBytecodeCache):
    pass


class MemcachedBytecodeCache(JinjaMemcachedBytecodeCache, AsyncBytecodeCache):
    pass



import os
import typing as t
import weakref
from inspect import iscoroutinefunction

from jinja2 import Environment, Template, Undefined
from jinja2.environment import internalcode
from jinja2.exceptions import (
    TemplateNotFound,
    TemplatesNotFound,
    TemplateSyntaxError,
    UndefinedError,
)

from .compiler import AsyncCodeGenerator

if t.TYPE_CHECKING:
    from .bccache import AsyncBytecodeCache
    from .loaders import AsyncBaseLoader


class AsyncEnvironment(Environment):
    code_generator_class: type["AsyncCodeGenerator"] = AsyncCodeGenerator
    bytecode_cache: t.Optional["AsyncBytecodeCache"]
    template_class: type["AsyncTemplate"]
    loader: t.Optional["AsyncBaseLoader"]

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        kwargs["enable_async"] = True
        super().__init__(*args, **kwargs)

    async def compile_templates_async(
        self,
        target: t.Union[str, "os.PathLike[str]"],
        extensions: t.Collection[str] | None = None,
        filter_func: t.Callable[[str], bool] | None = None,
        zip: str | None = "deflated",
        log_function: t.Callable[[str], None] | None = None,
        ignore_errors: bool = True,
    ) -> None:
        """Asynchronously finds all the templates the loader can find, compiles them
        and stores them in `target`.  If `zip` is `None`, instead of in a
        zipfile, the templates will be stored in a directory.
        By default, a deflate zip algorithm is used. To switch to
        the stored algorithm, `zip` can be set to ``'stored'``.

        `extensions` and `filter_func` are passed to :meth:`list_templates_async`.
        Each template returned will be compiled to the target folder or
        zipfile.

        By default, template compilation errors are ignored.  In case a
        log function is provided, errors are logged.  If you want template
        syntax errors to abort the compilation you can set `ignore_errors`
        to `False` and you will get an exception on syntax errors.
        """
        from jinja2.loaders import ModuleLoader

        if log_function is None:

            def log_function(x: str) -> None:
                pass

        assert log_function is not None
        assert self.loader is not None, "No loader configured."

        def write_file(filename: str, data: str) -> None:
            if zip:
                info = ZipInfo(filename)
                info.external_attr = 0o755 << 16
                zip_file.writestr(info, data)
            else:
                with open(os.path.join(target, filename), "wb") as f:
                    f.write(data.encode("utf8"))

        if zip is not None:
            from zipfile import ZIP_DEFLATED, ZIP_STORED, ZipFile, ZipInfo

            zip_file = ZipFile(
                target, "w", dict(deflated=ZIP_DEFLATED, stored=ZIP_STORED)[zip]
            )
            log_function(f"Compiling into Zip archive {target!r}")
        else:
            if not os.path.isdir(target):
                os.makedirs(target)
            log_function(f"Compiling into folder {target!r}")

        try:
            for name in await self.list_templates_async(extensions, filter_func):
                source, filename, _ = await self.loader.get_source_async(self, name)
                try:
                    code = self.compile(source, name, filename, True, True)
                except TemplateSyntaxError as e:
                    if not ignore_errors:
                        raise
                    log_function(f'Could not compile "{name}": {e}')
                    continue

                filename = ModuleLoader.get_module_filename(name)

                write_file(filename, code)
                log_function(f'Compiled "{name}" as {filename}')
        finally:
            if zip:
                zip_file.close()

        log_function("Finished compiling templates")

    async def list_templates_async(
        self,
        extensions: t.Collection[str] | None = None,
        filter_func: t.Callable[[str], bool] | None = None,
    ) -> list[str]:
        """Asynchronously returns a list of templates for this environment.
        This requires that the loader supports the loader's
        :meth:`~BaseLoader.list_templates_async` method.

        If there are other files in the template folder besides the
        actual templates, the returned list can be filtered.  There are two
        ways: either `extensions` is set to a list of file extensions for
        templates, or a `filter_func` can be provided which is a callable that
        is passed a template name and should return `True` if it should end up
        in the result list.

        If the loader does not support that, a :exc:`TypeError` is raised.
        """
        assert self.loader is not None, "No loader configured."
        names = await self.loader.list_templates_async()

        if extensions is not None:
            if filter_func is not None:
                raise TypeError(
                    "either extensions or filter_func can be passed, but not both"
                )

            def filter_func(x: str) -> bool:
                return "." in x and x.rsplit(".", 1)[1] in extensions

        if filter_func is not None:
            names = [name for name in names if filter_func(name)]

        return names

    @internalcode
    async def _load_template_async(
        self,
        name: str,
        globals: t.MutableMapping[str, t.Any] | None,
    ) -> "Template":
        if self.loader is None:
            raise TypeError("no loader for this environment specified")
        cache_key = (weakref.ref(self.loader), name)
        if self.cache is not None:
            template = self.cache.get(cache_key)
            if template is not None and (
                not self.auto_reload or await template.is_up_to_date_async
            ):
                if globals:
                    template.globals.update(globals)

                return template

        template = await self.loader.load_async(self, name, self.make_globals(globals))

        if self.cache is not None:
            self.cache[cache_key] = template
        return template

    @internalcode
    async def get_template_async(
        self,
        name: t.Union[str, "Template"],
        parent: str | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> "Template":
        """Asynchronously load a template by name with :attr:`loader` and return a
        :class:`Template`. If the template does not exist a
        :exc:`TemplateNotFound` exception is raised.

        :param name: Name of the template to load. When loading
            templates from the filesystem, "/" is used as the path
            separator, even on Windows.
        :param parent: The name of the parent template importing this
            template. :meth:`join_path` can be used to implement name
            transformations with this.
        :param globals: Extend the environment :attr:`globals` with
            these extra variables available for all renders of this
            template. If the template has already been loaded and
            cached, its globals are updated with any new items.
        """
        if isinstance(name, Template):
            return name
        if parent is not None:
            name = self.join_path(name, parent)

        return await self._load_template_async(name, globals)

    @internalcode
    async def select_template_async(
        self,
        names: t.Iterable[t.Union[str, "Template"]],
        parent: str | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> "Template":
        """Like :meth:`get_template_async`, but asynchronously tries loading
        multiple names. If none of the names can be loaded a
        :exc:`TemplatesNotFound` exception is raised.

        :param names: List of template names to try loading in order.
        :param parent: The name of the parent template importing this
            template. :meth:`join_path` can be used to implement name
            transformations with this.
        :param globals: Extend the environment :attr:`globals` with
            these extra variables available for all renders of this
            template. If the template has already been loaded and
            cached, its globals are updated with any new items.
        """
        if isinstance(names, Undefined):
            names._fail_with_undefined_error()

        if not names:
            raise TemplatesNotFound(
                message="Tried to select from an empty list of templates."
            )

        for name in names:
            if isinstance(name, Template):
                return name
            if parent is not None:
                name = self.join_path(name, parent)
            try:
                return await self._load_template_async(name, globals)
            except (TemplateNotFound, UndefinedError):
                pass
        raise TemplatesNotFound(names)  # type: ignore

    @internalcode
    async def get_or_select_template_async(
        self,
        template_name_or_list: t.Union[str, "Template", list[t.Union[str, "Template"]]],
        parent: str | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> "Template":
        """Use :meth:`select_template_async` if an iterable of template names
        is given, or :meth:`get_template_async` if one name is given.
        """
        if isinstance(template_name_or_list, (str, Undefined)):
            return await self.get_template_async(template_name_or_list, parent, globals)
        elif isinstance(template_name_or_list, Template):
            return template_name_or_list
        return await self.select_template_async(template_name_or_list, parent, globals)

    # Runtime aliases for async template loading. These intentionally change the
    # call contract compared to Environment (they return awaitables).
    get_template = get_template_async  # type: ignore[assignment]
    select_template = select_template_async  # type: ignore[assignment]
    get_or_select_template = get_or_select_template_async  # type: ignore[assignment]
    list_templates = list_templates_async  # type: ignore[assignment]


class AsyncTemplate(Template):
    environment_class: type[Environment] = AsyncEnvironment
    _uptodate: t.Callable[[], t.Awaitable[bool] | bool] | None

    @property
    async def is_up_to_date_async(self) -> bool:
        """If this variable is `False` there is a newer version available."""
        if self._uptodate is None:
            return True
        if iscoroutinefunction(self._uptodate):
            return await self._uptodate()
        return self._uptodate()


AsyncEnvironment.template_class = AsyncTemplate



from .bccache import AsyncBytecodeCache, FileSystemBytecodeCache, MemcachedBytecodeCache
from .compiler import AsyncCodeGenerator
from .environment import AsyncEnvironment, AsyncTemplate
from .loaders import (
    AsyncBaseLoader,
    ChoiceLoader,
    DictLoader,
    FileSystemLoader,
    FunctionLoader,
    ModuleLoader,
    PackageLoader,
    PrefixLoader,
)
from .sandbox import AsyncSandboxedEnvironment

__all__ = [
    "AsyncBytecodeCache",
    "FileSystemBytecodeCache",
    "MemcachedBytecodeCache",
    "AsyncCodeGenerator",
    "AsyncEnvironment",
    "AsyncTemplate",
    "AsyncBaseLoader",
    "ChoiceLoader",
    "DictLoader",
    "FileSystemLoader",
    "FunctionLoader",
    "ModuleLoader",
    "PackageLoader",
    "PrefixLoader",
    "AsyncSandboxedEnvironment",
]
